"""Useful (I guess) helpers."""
from yamkix import __version__


def get_version_string():
    """Return the version string."""
    return "yamkix v" + __version__


def print_version():
    """Print version."""
    print(get_version_string())


def remove_all_linebreaks(comment):
    """Remove trailing linebreak."""
    return comment.replace("\n", "")


def string_is_comment(a_string):
    """Is it a comment starting with a #."""
    return a_string[0] == "#"


def strip_leading_double_space(stream):
    """Strip the potential leading double spaces in CommentedSeq."""
    if stream.startswith("  "):
        stream = stream[2:]
    return stream.replace("\n  ", "\n")



"""Yamkix configuration helpers."""
import collections
import sys
from argparse import Namespace

from yamkix import __version__

YamkixConfig = collections.namedtuple(
    "YamkixConfig",
    "explicit_start \
        explicit_end \
        default_flow_style \
        dash_inwards \
        quotes_preserved \
        parsing_mode \
        spaces_before_comment \
        line_width \
        version \
        io_config",
)

YamkixInputOutputConfig = collections.namedtuple(
    "YamkixInputOutputConfig",
    "input input_display_name output output_display_name",
)


def get_default_yamkix_config() -> YamkixConfig:
    """Return Yamkix default config."""
    return YamkixConfig(
        parsing_mode="rt",
        explicit_start=True,
        explicit_end=False,
        default_flow_style=False,
        dash_inwards=True,
        quotes_preserved=True,
        spaces_before_comment=None,
        line_width=2048,
        version=False,
        io_config=get_default_yamkix_input_output_config(),
    )


# pylint: disable=too-many-arguments
def get_yamkix_config_from_default(
    parsing_mode=None,
    explicit_start=None,
    explicit_end=None,
    default_flow_style=None,
    dash_inwards=None,
    quotes_preserved=None,
    spaces_before_comment=None,
    line_width=None,
    io_config=None,
) -> YamkixConfig:
    """Return a Yamkix config based on default."""
    default_config = get_default_yamkix_config()
    return YamkixConfig(
        parsing_mode=parsing_mode if parsing_mode is not None else default_config.parsing_mode,
        explicit_start=explicit_start if explicit_start is not None else default_config.explicit_start,
        explicit_end=explicit_end if explicit_end is not None else default_config.explicit_end,
        default_flow_style=default_flow_style if default_flow_style is not None else default_config.default_flow_style,
        dash_inwards=dash_inwards if dash_inwards is not None else default_config.dash_inwards,
        quotes_preserved=quotes_preserved if quotes_preserved is not None else default_config.quotes_preserved,
        spaces_before_comment=spaces_before_comment
        if spaces_before_comment is not None
        else default_config.spaces_before_comment,
        line_width=line_width if line_width is not None else default_config.line_width,
        version=None,
        io_config=io_config,
    )


def get_default_yamkix_input_output_config() -> YamkixInputOutputConfig:
    """Return a default input / output config."""
    return YamkixInputOutputConfig(
        input=None,
        output=None,
        input_display_name="STDIN",
        output_display_name="STDOUT",
    )


def print_yamkix_config(yamkix_config: YamkixConfig):
    """Print a human readable Yamkix config on stderr."""
    yamkix_input_output_config = yamkix_config.io_config
    print(
        "[yamkix("
        + __version__
        + ")] Processing: input="
        + yamkix_input_output_config.input_display_name
        + ", output="
        + yamkix_input_output_config.output_display_name
        + ", typ="
        + yamkix_config.parsing_mode
        + ", explicit_start="
        + str(yamkix_config.explicit_start)
        + ", explicit_end="
        + str(yamkix_config.explicit_end)
        + ", default_flow_style="
        + str(yamkix_config.default_flow_style)
        + ", quotes_preserved="
        + str(yamkix_config.quotes_preserved)
        + ", dash_inwards="
        + str(yamkix_config.dash_inwards)
        + ", spaces_before_comment="
        + str(yamkix_config.spaces_before_comment)
        + ", show_version="
        + str(yamkix_config.version),
        file=sys.stderr,
    )


def get_input_output_config_from_args(
    args: Namespace,
) -> YamkixInputOutputConfig:
    """Get input, output and associated labels as YamkixInputOutputConfig."""
    input_display_name = "STDIN"
    if args.input is None:
        f_input = None
    else:
        f_input = args.input
        input_display_name = f_input
    if args.stdout:
        f_output = None
    else:
        if args.output is not None and args.output != "STDOUT":
            f_output = args.output
        else:
            if args.output == "STDOUT":
                f_output = None
            else:
                if f_input is None:
                    f_output = None
                else:
                    f_output = args.input
    if f_output is None:
        output_display_name = "STDOUT"
    else:
        output_display_name = f_output
    return YamkixInputOutputConfig(
        input=f_input,
        input_display_name=input_display_name,
        output=f_output,
        output_display_name=output_display_name,
    )


def get_config_from_args(args: Namespace, inc_io_config: bool = True) -> YamkixConfig:
    """Build a YamkixConfig object from parsed args."""
    if args.typ not in ["safe", "rt"]:
        raise ValueError(f"'{args.typ}' is not a valid value for option --typ. " "Allowed values are 'safe' and 'rt'")
    if inc_io_config:
        yamkix_input_output_config = get_input_output_config_from_args(args)
    else:
        yamkix_input_output_config = None
    default_yamkix_config = get_default_yamkix_config()
    return YamkixConfig(
        explicit_start=not args.no_explicit_start,
        explicit_end=args.explicit_end if args.explicit_end is not None else False,
        default_flow_style=args.default_flow_style if args.default_flow_style is not None else False,
        dash_inwards=not args.no_dash_inwards,
        quotes_preserved=not args.no_quotes_preserved,
        parsing_mode=args.typ,
        version=args.version if args.version is not None else False,
        spaces_before_comment=get_spaces_before_comment_from_args(args),
        io_config=yamkix_input_output_config,
        line_width=default_yamkix_config.line_width,
    )


def get_spaces_before_comment_from_args(args: Namespace):
    """Extract a valid value for spaces_before_comment from args."""
    if args.spaces_before_comment is None:
        spaces_before_comment = None
    else:
        try:
            spaces_before_comment = int(args.spaces_before_comment)
        except ValueError:
            spaces_before_comment = None
    return spaces_before_comment



"""Deal with comments."""
from ruamel.yaml.comments import CommentedBase, NotNone
from ruamel.yaml.error import CommentMark
from ruamel.yaml.tokens import CommentToken

from yamkix.helpers import remove_all_linebreaks, string_is_comment


def yamkix_add_eol_comment(self, comment, key=NotNone, column=None):
    """Provide a custom add_eol_comment function.

    We need to be able to tune the number of spaces between
    the content and the comment for CommentedSeqs and CommentedMaps
    see https://stackoverflow.com/q/60915926
    """
    # pylint: disable=protected-access
    org_col = column
    if column is None:
        try:
            column = self._yaml_get_column(key)
        except AttributeError:
            column = 0
    if comment[0] != "#":
        comment = "# " + comment
    if org_col != 0:  # only do this if the specified column is not the beginning of the line
        if comment[0] == "#":
            additional_spaces = 1 if org_col is None else org_col - 1
            comment = " " * additional_spaces + comment
            column = 0
    start_mark = CommentMark(column)
    comment_as_list = [CommentToken(comment, start_mark, None), None]
    self._yaml_add_eol_comment(comment_as_list, key=key)


CommentedBase.yaml_add_eol_comment = yamkix_add_eol_comment


def process_single_comment(data, comment, key, column):
    """Process a single comment."""
    comment = remove_all_linebreaks(comment)
    data.yaml_add_eol_comment(comment, key, column=column)


def fix_for_issue29(data, key):
    """Fix issue#29."""
    if data.ca.items[key][3] is not None:
        data.ca.items[key][3] = None


def process_comments(data, column=None):
    """Reposition comments."""
    if isinstance(data, dict):
        if data.ca and data.ca.items:
            # print("CommentedMap " + str(data.ca.items), file=sys.stderr)
            for key in data.ca.items.keys():
                if data.ca.items[key][2]:
                    comment = data.ca.items[key][2].value
                    fix_for_issue29(data, key)
                    if string_is_comment(comment):
                        process_single_comment(data, comment, key, column)
        for key, val in data.items():
            process_comments(key, column=column)
            process_comments(val, column=column)
    elif isinstance(data, list):
        if data.ca and data.ca.items:
            # print("CommentedSeq " + str(data.ca.items), file=sys.stderr)
            for key in data.ca.items.keys():
                if data.ca.items[key][0]:
                    comment = data.ca.items[key][0].value
                    if string_is_comment(comment):
                        process_single_comment(data, comment, key, column)
        for elem in data:
            process_comments(elem, column=column)



"""Deal with args."""

import argparse

from yamkix import __version__
from yamkix.config import get_config_from_args, YamkixConfig


def get_override_or_default(short_opt_override, key, default_value):
    """Return the override to apply or the default value."""
    if short_opt_override and key in short_opt_override:
        return short_opt_override[key]
    return default_value


def add_yamkix_options_to_parser(parser, short_opt_override=None):
    """Add yamkix reusable options to a parser object."""
    parser.add_argument(
        get_override_or_default(short_opt_override, "--typ", "-t"),
        "--typ",
        required=False,
        default="rt",
        help="the yaml parser mode. Can be `safe` or `rt`",
    )
    parser.add_argument(
        get_override_or_default(short_opt_override, "--no-explicit-start", "-n"),
        "--no-explicit-start",
        action="store_true",
        help="by default, explicit start of the yaml doc \
                                is `On`, you can disable it with this option",
    )
    parser.add_argument(
        get_override_or_default(short_opt_override, "--explicit-end", "-e"),
        "--explicit-end",
        action="store_true",
        help="by default, explicit end of the yaml doc \
                                is `Off`, you can enable it with this option",
    )
    parser.add_argument(
        get_override_or_default(short_opt_override, "--no-quotes-preserved", "-q"),
        "--no-quotes-preserved",
        action="store_true",
        help="by default, quotes are preserved \
                                you can disable this with this option",
    )
    parser.add_argument(
        get_override_or_default(short_opt_override, "--default-flow-style", "-f"),
        "--default-flow-style",
        action="store_true",
        help="enable the default flow style \
                                `Off` by default. In default flow style \
                                (with typ=`rt`), maps and lists are written \
                                like json",
    )
    parser.add_argument(
        get_override_or_default(short_opt_override, "--no-dash-inwards", "-d"),
        "--no-dash-inwards",
        action="store_true",
        help="by default, dash are pushed inwards \
                                use `--no-dash-inwards` to have the dash \
                                start at the sequence level",
    )
    parser.add_argument(
        get_override_or_default(short_opt_override, "--spaces-before-comment", "-c"),
        "--spaces-before-comment",
        default=None,
        help="specify the number of spaces between comments and content. \
                        If not specified, comments are left as is.",
    )


def build_parser():
    """Build the cli args parser."""
    parser = argparse.ArgumentParser(
        description=f"""Yamkix v{__version__}.
            Format yaml input file.
            By default, explicit_start is `On`, explicit_end is `Off`
            and array elements are pushed inwards the start of the
            matching sequence. Comments are preserved thanks to default
            parsing mode `rt`.
        """
    )
    parser.add_argument(
        "-i",
        "--input",
        required=False,
        help="""the file to parse or 'STDIN'.
            Defaults to 'STDIN' if not specified
        """,
    )
    parser.add_argument(
        "-o",
        "--output",
        required=False,
        help="the name of the file to generate (can be 'STDOUT') \
                            (same as input file if not specified, \
                                hence 'STDOUT' if 'STDIN' as input)",
    )
    parser.add_argument(
        "-s",
        "--stdout",
        action="store_true",
        help="output is STDOUT whatever the value for \
                        input (-i) and output (-o)",
    )
    add_yamkix_options_to_parser(parser)
    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="show yamkix version",
    )
    return parser


def parse_cli(args) -> YamkixConfig:
    """Parse the cli args."""
    parser = build_parser()
    args = parser.parse_args(args)
    yamkix_config = get_config_from_args(args, inc_io_config=True)
    return yamkix_config



"""Load a yaml file and save it formatted according to some rules."""
from __future__ import print_function

import os
import sys

from ruamel.yaml.scanner import ScannerError

from yamkix.args import parse_cli
from yamkix.comments import process_comments
from yamkix.config import print_yamkix_config, YamkixConfig
from yamkix.helpers import print_version, strip_leading_double_space
from yamkix.yaml_writer import get_opinionated_yaml_writer


def round_trip_and_format(yamkix_config: YamkixConfig):
    """Load a file and save it formated.

    FIXME
    """
    yaml = get_opinionated_yaml_writer(yamkix_config)
    yamkix_io_config = yamkix_config.io_config
    input_file = yamkix_io_config.input
    output_file = yamkix_io_config.output
    dash_inwards = yamkix_config.dash_inwards
    spaces_before_comment = yamkix_config.spaces_before_comment
    if input_file is not None:
        with open(input_file, mode="rt", encoding="UTF-8") as f_input:
            parsed = yaml.load_all(f_input.read())
    else:
        parsed = yaml.load_all(sys.stdin.read())
    ready_for_dump = []
    try:
        # Read the parsed content to force the scanner to issue errors if any
        for data in parsed:
            ready_for_dump.append(data)

    except ScannerError as scanner_error:
        print("Something is wrong in the input file, got error from Scanner")
        print(scanner_error)
        return
    yamkix_dump_all(ready_for_dump, yaml, dash_inwards, output_file, spaces_before_comment)


def yamkix_dump_all(one_or_more_items, yaml, dash_inwards, output_file, spaces_before_comment):
    """Dump all the documents from the input structure."""
    if output_file is None:
        out = sys.stdout

    # Clear the output file if it is a file and it exists
    if output_file is not None and os.path.isfile(output_file):
        with open(output_file, mode="w", encoding="UTF-8") as _:
            pass
    for doc in one_or_more_items:
        if output_file is None:
            yamkix_dump_one(doc, yaml, dash_inwards, out, spaces_before_comment)
        else:
            with open(output_file, mode="a", encoding="UTF-8") as out:
                yamkix_dump_one(doc, yaml, dash_inwards, out, spaces_before_comment)


def yamkix_dump_one(single_item, yaml, dash_inwards, out, spaces_before_comment):
    """Dump a single document."""
    if spaces_before_comment is not None:
        process_comments(single_item, column=spaces_before_comment)
    if dash_inwards and type(single_item).__name__ == "CommentedSeq":
        yaml.dump(single_item, out, transform=strip_leading_double_space)
    else:
        yaml.dump(single_item, out)


def main():
    """(re)format yaml."""
    yamkix_config = parse_cli(sys.argv[1:])
    if yamkix_config.version:
        print_version()
    else:
        print_yamkix_config(yamkix_config)
        round_trip_and_format(yamkix_config)


if __name__ == "__main__":
    main()



"""Helper to deal with Yamkix configuration of the YAML instance."""
from ruamel.yaml import YAML

from yamkix.config import YamkixConfig


def get_opinionated_yaml_writer(
    yamkix_config: YamkixConfig,
) -> YAML:
    """Configure a yaml parser/formatter the yamkix way.

    Args:
        yamkix_config: a YamkixConfig instance
    Returns:
        a ruamel.yaml YAML instance
    """
    yaml = YAML(typ=yamkix_config.parsing_mode)
    yaml.explicit_start = yamkix_config.explicit_start
    yaml.explicit_end = yamkix_config.explicit_end
    yaml.default_flow_style = yamkix_config.default_flow_style
    yaml.preserve_quotes = yamkix_config.quotes_preserved
    yaml.width = yamkix_config.line_width
    if yamkix_config.dash_inwards:
        yaml.indent(mapping=2, sequence=4, offset=2)
    return yaml



# -*- coding: utf-8 -*-
"""Allow yamkix to be executable through `python -m yamkix`."""
from __future__ import absolute_import

from .yamkix import main

if __name__ == "__main__":  # pragma: no cover
    main()
